import os

import django
from django.core import management

os.environ["DJANGO_SETTINGS_MODULE"] = "tests.settings"
django.setup()

management.call_command("makemigrations", "tests")
management.call_command("migrate")



class FieldMetaMixin:
    @classmethod
    def get_field_meta(cls, field_name):
        return cls._meta.get_field(field_name).get_meta()

    @classmethod
    def has_field_meta(cls, field_name):
        return hasattr(cls._meta.get_field(field_name), "_meta") \
               and cls.get_field_meta(field_name) is not None



class FieldWithMeta:
    def __init__(self, **kwargs):
        if 'meta' in kwargs:
            self._meta = kwargs['meta']
            del kwargs["meta"]
        super(self.__class__, self).__init__(**kwargs)

    def get_meta(self):
        return self._meta

    def deconstruct(self):
        """
        Makemigrations resolves the base classes of model fields using this.
        Without this, CharField() will be resolved
        to `model_field_meta.wrapper.CharField` instead of `django.db.models.fields.CharField`
        """
        base_class = self.__class__.__bases__[0]
        base_class_path = "{}.{}".format(base_class.__module__, base_class.__qualname__)
        (field_name, _, args, kwargs) = super(self.__class__, self).deconstruct()

        # The lines below are taken from Django source code: django.db.models.fields.Field.deconstruct()
        if base_class_path.startswith("django.db.models.fields.related"):
            base_class_path = base_class_path.replace("django.db.models.fields.related", "django.db.models")
        elif base_class_path.startswith("django.db.models.fields.files"):
            base_class_path = base_class_path.replace("django.db.models.fields.files", "django.db.models")
        elif base_class_path.startswith("django.db.models.fields.proxy"):
            base_class_path = base_class_path.replace("django.db.models.fields.proxy", "django.db.models")
        elif base_class_path.startswith("django.db.models.fields"):
            base_class_path = base_class_path.replace("django.db.models.fields", "django.db.models")

        return field_name, base_class_path, args, kwargs



from django.test import TestCase

from tests.models import Coffee


class FieldMetaMixinTestCase(TestCase):
    def setUp(self):
        Coffee.objects.all().delete()

    def _create_coffee(self):
        return Coffee.objects.create(brand="Malongo", blend="Arabica", packaging="0.5kg")

    def test_instance_get_meta(self):
        malongo_arabica = self._create_coffee()
        meta = malongo_arabica.get_field_meta("packaging")
        self.assertDictEqual(meta, Coffee.packaging_meta)

    def test_class_get_meta(self):
        meta = Coffee.get_field_meta("packaging")
        self.assertDictEqual(meta, Coffee.packaging_meta)

    def test_instance_has_meta(self):
        malongo_arabica = self._create_coffee()
        self.assertFalse(malongo_arabica.has_field_meta("brand"))
        self.assertFalse(malongo_arabica.has_field_meta("blend"))
        self.assertTrue(malongo_arabica.has_field_meta("packaging"))

    def test_class_has_meta(self):
        self.assertFalse(Coffee.has_field_meta("brand"))
        self.assertFalse(Coffee.has_field_meta("blend"))
        self.assertTrue(Coffee.has_field_meta("packaging"))



from typing import Iterable

from django.db import models as django_models
from django.db.models.fields import __all__ as django_fields

from model_field_meta.fields import FieldWithMeta


class WrappedDjangoModels:
    new_classes = {}

    def __init__(self):
        # Model fields
        for field in django_fields:
            if field.endswith("Field") and field != "Field":
                base_class = getattr(django_models, field)
                self.new_classes[field] = type(field, (base_class,), {
                    "__init__": FieldWithMeta.__init__,
                    "_meta": None,
                    "get_meta": FieldWithMeta.get_meta,
                    "deconstruct": FieldWithMeta.deconstruct,
                    "__module__": base_class.__module__,
                })

    def __getattr__(self, attr):
        if attr in self.new_classes:
            return self.new_classes[attr]
        return getattr(django_models, attr)

    def __dir__(self) -> Iterable[str]:
        return django_fields



from model_field_meta.mixins import FieldMetaMixin
from model_field_meta.wrapper import WrappedDjangoModels

models = WrappedDjangoModels()

__all__ = ["models", "FieldMetaMixin"]
__version__ = "0.5"



import tempfile

TEMP_DIR = tempfile.TemporaryDirectory().name

INSTALLED_APPS = (
    'model_field_meta',
    'tests',
)

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': TEMP_DIR,
    }
}

SECRET_KEY = 'THIS IS A SECRET KEY'



"""
Test the migration files generated
"""
import glob

from django.core import management
from django.test import TestCase


class MigrationFileTestSuite(TestCase):
    def test_migrations_are_not_regenerated_unintentionally(self):
        migration_files = glob.glob("./tests/migrations/00*.*")
        self.assertEqual(len(migration_files), 1)

        for _ in range(1, 5):
            management.call_command("makemigrations", "tests")

        migration_files = glob.glob("./tests/migrations/00*.*")
        self.assertEqual(len(migration_files), 1)



from django.test import TestCase

from model_field_meta import models


class FieldMetaTestSuite(TestCase):
    def test_empty_meta(self):
        field = models.TextField()
        self.assertFalse(field.get_meta())

    def test_simple_meta(self):
        field = models.TextField(meta="a simple value")
        self.assertEqual(field.get_meta(), "a simple value")

    def test_complex_meta(self):
        class Options:
            food_options = (
                ("1", "Pizza"),
                ("2", "Sushi"),
                ("3", "Ratatouille"),
            )

        field = models.TextField(meta={"options": Options.food_options})
        self.assertDictEqual(field.get_meta(), {"options": Options.food_options})



from model_field_meta import models, FieldMetaMixin


class Coffee(FieldMetaMixin, models.Model):
    packaging_meta = {"choices": {
        "kg": (0.5, 1, 2, 5, 10, 15, 20),
    }}

    brand = models.CharField(max_length=512)
    blend = models.CharField(max_length=512, meta=None)
    packaging = models.TextField(meta=packaging_meta)
